@@ -4,7 +4,8 @@ import io
 import boto3
 from io import StringIO
 from multiprocessing.connection import wait
-from multiprocessing import Process, Pipe
+from multiprocessing import Process, Pipe, Queue
+from queue import Empty
 import pandas as pd
 from pandas import DataFrame
 from tqdm import tqdm
@@ -32,6 +33,9 @@ TRANSFORM_ONE_OBJ = "infin_transform_one_object"
 
 verbose = False
 
+def num_threads():
+    return 8
+
 class FuncLister(ast.NodeVisitor):
     def __init__(self, glbs):
         self.glbs = glbs;
@@ -116,80 +120,53 @@ def list_dir_by_dir(client, bucket, prefix_in, recurse, dict_of_arrays_of_files)
                 if (bool(recurse) and this_prefix != None ):
                     list_dir_by_dir(client, bucket, this_prefix, recurse, dict_of_arrays_of_files)
 
-def num_threads():
-    return 8
-
 def info(msg):
     now = datetime.now()
     print(__name__ + '[' + str(os.getpid()) + '][' + now.strftime('%Y-%m-%d %H:%M:%S') + ']' + msg)
 
-def s3downloader(write_pipe, read_pipe, endpoint, glb):
+class download_task:
+    def __init__(self, command, bucketname, parentdir, filename):
+        self.command = command
+        self.bucketname = bucketname
+        self.parentdir = parentdir
+        self.filename = filename
+
+def s3downloader(download_task_q, write_pipe, endpoint, glb):
     # info('s3downloader: started. endpoint=' + endpoint)
 
     if (glb != None):
-        namespaced_infin_transform_one_object = namespaced_function(glb['infin_transform_one_object'], glb, None, True)
-
-    session = boto3.session.Session(profile_name='infinstor')
-    client = session.client('s3', endpoint_url=endpoint)
-    while (True):
-        op = read_pipe.recv()
-        if (op == 'download'):
-            bucketname = read_pipe.recv()
-            filename = read_pipe.recv()
-            # info('s3downloader: starting download of ' + filename + ' from ' + bucketname)
-            obj = client.get_object(Bucket=bucketname, Key=filename)
-            # info('s3downloader: finished download of ' + filename + ' from ' + bucketname)
-            strbody = obj['Body']
-            datetime = obj['LastModified']
-            bts = strbody.read()
-            key = datetime.strftime('%Y-%m-%d %H:%M:%S') + ' ' + bucketname + '/' + filename
-            if (glb == None):
-                write_pipe.send(key)
-                write_pipe.send(bts)
-            else:
-                try:
-                    namespaced_infin_transform_one_object(bucketname, filename, bts)
-                except Exception as e:
-                    status_str = str(e)
-                    info("Error executing infin_transform_one_object: " + status_str)
-                else:
-                    status_str = 'Success'
-                write_pipe.send(key)
-                write_pipe.send(status_str)
-        elif (op == 'quit'):
-            break
-
-def s3downloader_by_dir(write_pipe, read_pipe, endpoint, glb):
-    # info('s3downloader_by_dir: started. endpoint=' + endpoint)
-
-    if (glb != None):
-        namespaced_infin_transform_one_object = \
+        namespaced_infin_transform_one_object =\
                 namespaced_function(glb['infin_transform_one_object'], glb, None, True)
 
     session = boto3.session.Session(profile_name='infinstor')
     client = session.client('s3', endpoint_url=endpoint)
     while (True):
-        op = read_pipe.recv()
+        try:
+            download_task = download_task_q.get()
+        except Empty as e:
+            info('s3downloader: No more entries in download_task_q. Exiting..')
+            write_pipe.close()
+            break
+        op = download_task.command
         if (op == 'download'):
-            bucketname = read_pipe.recv()
-            parentdir = read_pipe.recv()
-            filename = read_pipe.recv()
-            # info('s3downloader_by_dir: starting download ' + filename + ' from ' + bucketname + ' parent=' + parentdir)
+            bucketname = download_task.bucketname
+            parentdir = download_task.parentdir
+            filename = download_task.filename
             if (parentdir == ''):
                 full_object_key = filename
             else:
                 full_object_key = parentdir + sep + filename
+            # info('s3downloader: starting download of ' + full_object_key + ' from ' + bucketname)
             obj = client.get_object(Bucket=bucketname, Key=full_object_key)
-            # info('s3downloader_by_dir: finished download ' + filename + ' from ' + bucketname + ' parent=' + parentdir)
+            # info('s3downloader: finished download of ' + full_object_key + ' from ' + bucketname)
             strbody = obj['Body']
             datetime = obj['LastModified']
             bts = strbody.read()
-            key = datetime.strftime('%Y-%m-%d %H:%M:%S') + ' ' + bucketname + '/' + filename
+            key = datetime.strftime('%Y-%m-%d %H:%M:%S') + ' ' + bucketname + '/' + full_object_key
             if (glb == None):
                 write_pipe.send(key)
                 write_pipe.send(bts)
             else:
-                # info("s3downloader_by_dir: Executing infin_transform_one_object for " + filename)
                 try:
                     namespaced_infin_transform_one_object(bucketname, parentdir, filename, bts)
                 except Exception as e:
@@ -197,10 +174,15 @@ def s3downloader_by_dir(write_pipe, read_pipe, endpoint, glb):
                     info("Error executing infin_transform_one_object: " + status_str)
                 else:
                     status_str = 'Success'
-                # info("s3downloader_by_dir: Finshed infin_transform_one_object for " + filename)
                 write_pipe.send(key)
                 write_pipe.send(status_str)
         elif (op == 'quit'):
+            # info('s3downloader: Received quit')
+            write_pipe.close()
+            break
+        else:
+            info('s3downloader: Unknown command ' + op)
+            write_pipe.close()
             break
 
 def namespaced_function(function, global_dict, defaults=None, preserve_context=False):
@@ -233,96 +215,55 @@ def load_one_csv_from_bytearray(bts):
     sio = StringIO(s)
     return pd.read_csv(sio)
 
-def download_objects(endpoint, bucketname, array_of_files, is_csv, glb):
+def download_objects(endpoint, bucketname, parentdir, array_of_files, is_csv, glb):
     try:
         start_run()
-        objects = download_objects_inner(endpoint, bucketname, array_of_files, is_csv, None)
+        return download_objects_inner(endpoint, bucketname, parentdir, array_of_files, is_csv, None)
     finally:
         end_run()
 
 # returns a pandas DataFrame with index 'YY-MM-dd HH:MM:SS bucketname/filename'
 # and one column named RawBytes that contains the raw bytes from the object
 # This inner method is called after the outer method wraps in mlflow start_run
-def download_objects_inner(endpoint, bucketname, array_of_files, is_csv, glb):
+def download_objects_inner(endpoint, bucketname, parentdir, array_of_files, is_csv, glb):
     log_param("object_count", len(array_of_files))
 
-    filebytes = []
-    filekeys = []
-    toChildProcess = []
-    fromChildProcess = []
+    command_q = Queue(len(array_of_files) + num_threads())
+    for onefile in array_of_files:
+        command_q.put(download_task('download', bucketname, parentdir, onefile))
+    for i in range(num_threads()):
+        command_q.put(download_task('quit', '', '', ''))
+
+    pipe_from_child = []
     processes = []
-    available_processes = []
     for i in range(num_threads()):
         r1, w1 = Pipe()
-        fromChildProcess.append(r1)
-        r2, w2 = Pipe()
-        toChildProcess.append(w2) 
-        p = Process(target=s3downloader, args=(w1, r2, endpoint, glb,))
+        pipe_from_child.append(r1)
+        p = Process(target=s3downloader, args=(command_q, w1, endpoint, glb,))
         p.start()
         processes.append(p)
-        available_processes.append(True)
         w1.close()
-        r2.close()
 
+    filebytes = []
+    filekeys = []
+    step = 0
     with tqdm(total=len(array_of_files)) as pbar:
-        read_pipes = []
-        next_file = 0
-        step = 0
-        while (True):
-            if (next_file == len(array_of_files)):
-                break
-            for i in range(num_threads()):
-                if (available_processes[i] == True):
-                    available_processes[i] = False
-                    toChildProcess[i].send('download')
-                    toChildProcess[i].send(bucketname)
-                    toChildProcess[i].send(array_of_files[next_file])
-                    read_pipes.append(fromChildProcess[i])
-                    next_file += 1
-                    if (next_file == len(array_of_files)):
-                        break
-                    pbar.update(1)
-                    if ((next_file % 10) == 0):
-                        log_metric("downloaded", next_file, step=step)
-                        step += 1
-
-            if (len(read_pipes) > 0):
-                for read_pipe in wait(read_pipes):
-                    try:
-                        key = read_pipe.recv()
-                        bts = read_pipe.recv()
-                        for i in range(num_threads()):
-                            if (fromChildProcess[i] == read_pipe):
-                                available_processes[i] = True
-                                break
-                        filekeys.append(key)
-                        filebytes.append(bts)
-                        break
-                    except EOFError:
-                        info('eof received')
-                    finally:
-                        read_pipes.remove(read_pipe)
-
-        while (len(read_pipes) > 0):
-            for read_pipe in wait(read_pipes):
+        files_read = 0
+        while (len(pipe_from_child) > 0):
+            for read_pipe in wait(pipe_from_child):
                 try:
                     key = read_pipe.recv()
                     bts = read_pipe.recv()
-                    for i in range(num_threads()):
-                        if (fromChildProcess[i] == read_pipe):
-                            available_processes[i] = True
-                            break
                     filekeys.append(key)
                     filebytes.append(bts)
-                    continue
+                    pbar.update(1)
+                    files_read += 1
+                    if ((files_read % 10) == 0):
+                        log_metric("downloaded", files_read, step=step)
+                        step += 1
                 except EOFError:
-                    info('eof received')
-                finally:
-                    read_pipes.remove(read_pipe)
-        print('finished reading all files')
+                    pipe_from_child.remove(read_pipe)
 
-    for i in range(num_threads()):
-        toChildProcess[i].send('quit')
     for i in range(num_threads()):
         processes[i].join()
     if (is_csv == True):
@@ -330,100 +271,7 @@ def download_objects_inner(endpoint, bucketname, array_of_files, is_csv, glb):
     else:
         data = {'RawBytes': filebytes}
         rv = DataFrame(data, index=filekeys)
-    log_metric("downloaded", next_file, step=step)
-    return rv
-
-# returns a pandas DataFrame with index 'YY-MM-dd HH:MM:SS bucketname/filename'
-# and one column named RawBytes that contains the raw bytes from the object
-def download_objects_by_dir(endpoint, bucketname, parentdir, array_of_files, is_csv, glb):
-    log_param(parentdir, len(array_of_files))
-
-    filebytes = []
-    filekeys = []
-    toChildProcess = []
-    fromChildProcess = []
-    processes = []
-    available_processes = []
-    for i in range(num_threads()):
-        r1, w1 = Pipe()
-        fromChildProcess.append(r1)
-        r2, w2 = Pipe()
-        toChildProcess.append(w2) 
-        p = Process(target=s3downloader_by_dir, args=(w1, r2, endpoint, glb,))
-        p.start()
-        processes.append(p)
-        available_processes.append(True)
-        w1.close()
-        r2.close()
-
-    with tqdm(total=len(array_of_files)) as pbar:
-        read_pipes = []
-        next_file = 0
-        step = 0
-        while (True):
-            if (next_file == len(array_of_files)):
-                break
-            for i in range(num_threads()):
-                if (available_processes[i] == True):
-                    available_processes[i] = False
-                    toChildProcess[i].send('download')
-                    toChildProcess[i].send(bucketname)
-                    toChildProcess[i].send(parentdir)
-                    toChildProcess[i].send(array_of_files[next_file])
-                    read_pipes.append(fromChildProcess[i])
-                    next_file += 1
-                    if (next_file == len(array_of_files)):
-                        break
-                    pbar.update(1)
-                    if ((next_file % 10) == 0):
-                        log_metric("downloaded", next_file, step=step)
-                        step += 1
-
-            if (len(read_pipes) > 0):
-                for read_pipe in wait(read_pipes):
-                    try:
-                        key = read_pipe.recv()
-                        bts = read_pipe.recv()
-                        for i in range(num_threads()):
-                            if (fromChildProcess[i] == read_pipe):
-                                available_processes[i] = True
-                                break
-                        filekeys.append(key)
-                        filebytes.append(bts)
-                        break
-                    except EOFError:
-                        info('eof received')
-                    finally:
-                        read_pipes.remove(read_pipe)
-
-        while (len(read_pipes) > 0):
-            for read_pipe in wait(read_pipes):
-                try:
-                    key = read_pipe.recv()
-                    bts = read_pipe.recv()
-                    for i in range(num_threads()):
-                        if (fromChildProcess[i] == read_pipe):
-                            available_processes[i] = True
-                            break
-                    filekeys.append(key)
-                    filebytes.append(bts)
-                    continue
-                except EOFError:
-                    info('eof received')
-                finally:
-                    read_pipes.remove(read_pipe)
-        print('finished reading all files')
-
-    for i in range(num_threads()):
-        toChildProcess[i].send('quit')
-    for i in range(num_threads()):
-        processes[i].join()
-    if (is_csv == True):
-        rv = pd.concat(map(load_one_csv_from_bytearray, filebytes))
-    else:
-        data = {'RawBytes': filebytes}
-        rv = DataFrame(data, index=filekeys)
-    log_metric("downloaded", next_file, step=step)
+    log_metric("downloaded", files_read, step=step)
     return rv
 
 def actually_run_transformation(client, endpoint, cache_path, is_pandas_df, bucketname,\
@@ -439,7 +287,7 @@ def actually_run_transformation(client, endpoint, cache_path, is_pandas_df, buck
     print('Total Number Of Objects: ' + str(len(array_of_files)))
     try:
         start_run()
-        objects = download_objects_inner(endpoint, bucketname, array_of_files, is_csv, None)
+        objects = download_objects_inner(endpoint, bucketname, '', array_of_files, is_csv, None)
     finally:
         end_run()
 
@@ -576,7 +424,8 @@ def read_raw_and_xform_one_object(time_spec, service_name, bucketname,\
             # pprint.pprint(glb)
             info("Execution of global statics for parentdir " + parentdir\
                     + " complete. status_str=" + status_str)
-            objects = download_objects_by_dir(endpoint, bucketname, parentdir, array_of_files, False, glb)
+            objects = download_objects_inner(endpoint, bucketname, parentdir, array_of_files,\
+                    False, glb)
     finally:
         end_run()
     return objects
